use std::{collections::HashMap, path::Path};

use uuid::Uuid;

use crate::{playlist::Playlist, song::Song};

mod json;
mod toml;
pub use json::JsonStore;
pub use toml::TomlStore;

pub trait BaseStore { 
    fn get_songs(&self) -> &HashMap<Uuid, Song>;
    fn get_songs_mut(&mut self) -> &mut HashMap<Uuid, Song>;
    fn get_song(&self, id: &Uuid) -> Option<&Song>;
    fn get_song_mut(&mut self, id: &Uuid) -> Option<&mut Song>;

    fn get_playlists(&self) -> &HashMap<Uuid, Playlist>;
    fn get_playlists_mut(&mut self) -> &mut HashMap<Uuid, Playlist>;
    fn get_playlist(&self, id: &Uuid) -> Option<&Playlist>;
    fn get_playlist_mut(&mut self, id: &Uuid) -> Option<&mut Playlist>;

    fn to_bytes(&self) -> crate::Result<Vec<u8>>;
    fn from_bytes(s: &[u8]) -> crate::Result<Self> where Self: Sized;
    fn empty() -> Self where Self: Sized;
    fn save_to(&self, p: &Path) -> crate::Result<()> {
        let bin = self.to_bytes()?;
        std::fs::write(p, bin)?;
        Ok(())
    }
    fn load_from(&mut self, p: &Path) -> crate::Result<()> where Self: Clone {
        let bin = std::fs::read(p)?;
        let s = Self::from_bytes(&bin)?;
        self.clone_from(&s);
        Ok(())
    }
    fn get_default_file_contents() -> &'static str;
    fn get_file_extension() -> &'static str;
    fn save_original_path(&mut self, p: &Path);
    fn get_original_path(&self) -> &Path;
}

pub trait StoreExtras: BaseStore {
    fn get_songs_sorted(&self) -> Vec<(&Uuid, &Song)> {
        let mut songs: Vec<_> = self.get_songs().iter().collect();
        songs.sort_by(|a, b| {
            let a = a.1.name().to_lowercase();
            let b = b.1.name().to_lowercase();
            a.cmp(&b)
        });
        songs
    }

    fn get_songs_sorted_cloned(&self) -> Vec<(Uuid, Song)> {
        let mut songs = Vec::new();
        for song in self.get_songs_sorted() {
            songs.push((song.0.clone(), song.1.clone()));
        }
        songs
    }

    fn get_playlists_sorted(&self) -> Vec<(&Uuid, &Playlist)> {
        let mut songs: Vec<_> = self.get_playlists().iter().collect();
        songs.sort_by(|a, b| {
            let a = a.1.name().to_lowercase();
            let b = b.1.name().to_lowercase();
            a.cmp(&b)
        });
        songs
    }
    fn get_playlist_songs_sorted(&self, pid: uuid::Uuid) -> Option<Vec<(&Uuid, &Song)>> {
        let songs_ids: Vec<_> = self.get_playlist(&pid)?.songs().iter().collect();
        let mut songs = Vec::new();

        for sid in songs_ids {
            songs.push((sid, self.get_song(sid)?));
        }

        songs.sort_by(|a, b| {
            let a = a.1.name().to_lowercase();
            let b = b.1.name().to_lowercase();
            a.cmp(&b)
        });
        Some(songs)
    }
    fn get_playlist_songs_sorted_cloned(&self, pid: uuid::Uuid) -> Option<Vec<(Uuid, Song)>> {
        let mut songs = Vec::new();
        for song in self.get_playlist_songs_sorted(pid)? {
            songs.push((song.0.clone(), song.1.clone()));
        }
        Some(songs)
    }
}

